
from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "batorch",
    fileinfo = "The inherited tensor from 'torch' with batch. This is the auto-generator, run the file to automatically generate tensor.py. Please DONOT touch tensor.py, modify tensor_template.py instead. ",
    requires = "torch"
)

from tensorsize import *
from device import AutoDevice
from collections import defaultdict
_device = AutoDevice(verbose=True, always_proceed=True)

with __info__:
    import torch
    from pycamia import scope, alias
    from pycamia import Path, SPrint
    from pycamia import tokenize, token_replace, identity_function
    from pycamia import python_lines, add_lineno
    from pycamia import get_num_indent, tab_to_indent, no_indent, with_indent, get_indent_len
    
with open(Path(__file__).parent / 'tensor_template.py') as fp:
    __text__ = fp.read()
    
    func_lines = []
    start_record_func = False
    size_mapping_line = None
    size_mapping_op_line = None
    for line in python_lines(__text__):
        if get_num_indent(line) == 0: start_record_func = False
        if start_record_func: func_lines.append(line)
        if line.lstrip().startswith("size_mapping ="): size_mapping_line = no_indent(line)
        if line.lstrip().startswith("size_mapping_op ="): size_mapping_op_line = no_indent(line)
        if line.lstrip().startswith("def matmul_shape"): start_record_func = True; func_lines = [line]
    exec(size_mapping_line)
    exec('\n'.join(func_lines))
    exec(size_mapping_op_line)

def template(decorators, _def, _fname, _args, doc, cast, get_size, inner_codes, manual_param):
    return f'''
{decorators}
{_def} {_fname}({_args}):
    """
{doc}
    """
    # Cast all arguments to the wanted 'Tensor' or 'Size' type, and convert the dim arguments. 
{cast}
    # Obtain available sized in arguments (which will be fed into size function).  
{get_size}
    # Use the given inner codes if they are provided. 
{inner_codes}
{manual_param}
    if getattr(obj, 'grad_fn', None) is not None:
        obj.grad_fn_name = "{_fname}"
    return obj
    '''

def get_updated_code(codes, mode='function', ignore_args=[]):
    """
    Get codes for inheritance.
    
    Args:
        func (function): the function object to be expanded. 
        mode (str: function|method): the string indicating the block is a function or a method. 
        ignore_arg (list): the list of arguments that are ignored in auto generated codes.
    """
    
    line_no = 0
    func_lines = python_lines(tab_to_indent(codes))
    declaration = func_lines[line_no]
    decorators = []
    while declaration.startswith('@'):
        decorators.append(declaration)
        line_no += 1
        declaration = func_lines[line_no].strip()
    decorators = '\n'.join(decorators)
    
    # Interprete the declaration line of template. 
    _def, _fname, _args, *_tail = tokenize(declaration, sep=[' ', '(', ')', '\n'])
    is_inplace = _fname.endswith('_') and _fname[-2:] != '__'
    line_no += 1
        
    line0 = tokenize(declaration, sep=':', by="()[]{}$$``''\"\"##")[-1].strip()
        
    # Extract the document string from the template. 
    doc_in_template = ''
    if not line0:
        line = func_lines[line_no]
        line_start = line.split('\n', 1)[0].lstrip("\t rf")
        if line_start.startswith('"""') or line_start.startswith("'''") or line_start.startswith('#'):
            doc_in_template = line.strip('\t \'"').strip('\n')
            line_no += 1
    
    # Copy the inner codes from the template. 
    num_indent = get_num_indent(declaration)
    indent = lambda k: " " * get_indent_len() * k
    inner_codes = ""
    if not line0 in ('...', 'pass'):
        inner_code_lines = ([with_indent(line0, 1)] if line0 else []) + func_lines[line_no:]
        for i, inner_line in enumerate(inner_code_lines):
            nosp_line = inner_line.lstrip(' ')
            prefix = ' ' * (len(inner_line) - len(nosp_line))
            nosp_line = ';'.join(f'obj = {x[7:]}' if x.lstrip().startswith('return ') else x for x in tokenize(nosp_line, sep=';'))
            nosp_line = ':'.join(f'obj = {x[7:]}' if x.lstrip().startswith('return ') else x for x in tokenize(nosp_line, sep=':'))
            inner_code_lines[i] = prefix + nosp_line
        inner_codes = '\n'.join(inner_code_lines) + '\n'
    if inner_codes.strip() in ('...', 'pass'): inner_codes = ""
    
    # Parse arguments. 
    for ig in ignore_args:
        _args = token_replace(_args, lambda x: x.lstrip('*').startswith(ig), '', sep=[','], strip=' ').replace(',,', ',').rstrip(',')
    
    args_name = None
    kwargs_name = None
    func_args = []
    for part in tokenize(_args, ','):
        default_value = None
        annotation = None
        seq = tokenize(part, '=')
        if len(seq) == 1: var_name = seq[0]
        elif len(seq) == 2: part, default_value = seq
        seq = tokenize(part, ':')
        if len(seq) == 1: var_name = seq[0]
        elif len(seq) == 2: var_name, annotation = seq
        if var_name: var_name = var_name.strip()
        if annotation: annotation = annotation.strip()
        if default_value: default_value = default_value.strip()
        if var_name in "*/": continue
        if var_name.startswith('**'): kwargs_name = var_name.lstrip('*')
        elif var_name.startswith('*'): args_name = var_name.lstrip('*')
        func_args.append((var_name.lstrip('*'), annotation, default_value))
    
    # Classify the arguments. 
    self_name = func_args[0][0]
    
    tensor_args = []
    size_args = []
    dim_args = []
    iter_dim_args = []
    rev_dim_args = []
    rmv_star_args = []
    
    inherit_args = []
    inherit_kwargs = []
    
    for name, annot, default in func_args:
        if annot is None: annot = ''
        if annot.endswith('[:]'): dim_args.append(name); rmv_star_args.append(name)
        elif annot.endswith(']') and 'dim[' in annot:
            dim_args.append(name)
            iter_dim_args.append(name)
            if annot.startswith('del_dim['):
                rev_dim_args.append(name)
        elif annot.strip('\'"') == 'Tensor': tensor_args.append(name)
        elif annot.strip('\'"') == 'Size': size_args.append(name)
        elif annot in "new_dim del_dim exist_dim".split(): dim_args.append(name)
        elif annot.startswith('linalg_dim'): dim_args.append(name)
        
        if name == args_name: inherit_kwargs.append('*' + name)
        elif name == kwargs_name: inherit_kwargs.append('**' + name)
        elif default is not None: inherit_kwargs.append(f"{name}={name}")
        elif mode == 'function' or name != self_name: inherit_args.append(name)
    inherit_args = ','.join(inherit_args)
    inherit_kwargs = ','.join(inherit_kwargs)
    
    # Generate casting codes. 
    annot_dict = {n: a for n, a, d in func_args}
    cast = with_indent('\n'.join(
        ["pivot = None", 
        f"for t in [{', '.join(tensor_args)}]:",
         with_indent("if isinstance(t, torch.Tensor): pivot = t", 1), 
         "subclass = Tensor.get_tensor_subclass(pivot)"] + 
        ['\n'.join([
            f"try: {x} = torch.tensor({x}) if pivot is None else tensor_to({x}, pivot) if {x} is not None and not isinstance({x}, torch.Tensor) else {x}",
             "except TypeError: return NotImplemented", 
            f"{x} = {x}.as_subclass(subclass).special_from({x}.shape) if {x} is not None and not isinstance({x}, subclass) else {x}",
            f"if {x}.device != {tensor_args[0]}.device and {x}.device.type == 'cpu': {x} = {x}.to({tensor_args[0]}.device)" if i > 0 else '',
        ]) for i, x in enumerate(tensor_args)] + 
        [f"{x} = {annot_dict[x]}({self_name}, %s{x})"%('*' if x == args_name else '') for x in dim_args]
    ), 1)
    
    get_size = with_indent('\n'.join([f"{x}_shape=None if {x} is None else Size({x}.shape)" for x in tensor_args] + [f'{x}=Size(%s{x})'%('*' if x == args_name else '') for x in size_args]), 1)
    size_fname = _fname
    if is_inplace: size_fname = size_fname.rstrip('_')
    size_arguments = ', '.join([f'{x}_shape' for x in tensor_args] + size_args + dim_args)
    
    # Auto generate auto 'inner_codes' for cases where no inner-codes are provided. 
    kwargs_dict = f"dict({', '.join(f'{n}={n}' for n, a, d in func_args if d is not None)})"
    size_map_func = 'size_mapping_op' if len(tensor_args) >= 2 else 'size_mapping'
    kwarg = f", **{kwargs_dict}" if eval(size_map_func)[size_fname].__code__.co_flags & 0x08 else ''
    ref_shape_defined = False
    if not inner_codes:
        if len(tensor_args) >= 2:
            inner_codes = indent(1) + "ref_shape, " + ', '.join([f'{x}_shape' for x in tensor_args]) + \
                f" = {size_map_func}['{size_fname}']({size_arguments}{kwarg})\n"
            for x in tensor_args: inner_codes += indent(1) + f"{x} = {x}.view({x}_shape)\n"
            ref_shape_defined = True
        
        parent = f"torch_super({self_name}, '{_fname}')" if mode == 'method' else f'torch.{_fname}'
        if len(iter_dim_args) > 0:
            lists = ', '.join([iarg + '[::-1]' if iarg in rev_dim_args else iarg for iarg in iter_dim_args])
            iters = ', '.join([d + '_iter' for d in iter_dim_args])
            if len(iter_dim_args) == 1: iters += ','
            for d in iter_dim_args:
                inherit_args = token_replace(inherit_args, d, d + "_iter", sep=[',', '='], strip=' ')
                inherit_kwargs = token_replace(inherit_kwargs, f"{d}={d}", f"{d}={d}_iter", sep=[','], strip=' ')
                if d == args_name:
                    inherit_kwargs = token_replace(inherit_kwargs, lambda x: x == '*' + args_name, d+'_iter', sep=[',', '='], strip=' ')
            inner_codes += indent(1) + "with torch._C.DisableTorchFunction():\n"
            inner_codes += indent(2) + f"for {iters} in zip({lists}):\n"
            inner_codes += indent(3) + f"auto_generated_args = tuple(x for x in [{inherit_args}] if x is not None)\n"
            if is_inplace: inner_codes += indent(3) + f"{parent}(*auto_generated_args, {inherit_kwargs})\n"
            else: inner_codes += indent(3) + f"{self_name} = {parent}(*auto_generated_args, {inherit_kwargs})\n"
            inner_codes += indent(1) + f"obj = {self_name}\n"
        else:
            for d in rmv_star_args:
                inherit_kwargs = token_replace(inherit_kwargs, lambda x: x.lstrip('*') == d, d, sep=[',', '='], strip=' ')
            inner_codes += indent(1) + "with torch._C.DisableTorchFunction():\n"
            inner_codes += indent(2) + f"auto_generated_args = tuple(x for x in [{inherit_args}] if x is not None)\n"
            if is_inplace:
                inner_codes += indent(2) + f"{parent}(*auto_generated_args, {inherit_kwargs})\n"
                inner_codes += indent(2) + f"obj = {self_name}\n"
            else: inner_codes += indent(2) + f"obj = {parent}(*auto_generated_args, {inherit_kwargs})\n"   
    if 'special_from' not in inner_codes and 'Tensor.inherit_from' not in inner_codes:
        if not ref_shape_defined:
            inner_codes += indent(1) + f"ref_shape{', *_' if size_map_func == 'size_mapping_op' else ''} = {size_map_func}['{size_fname}']({size_arguments}{kwarg})\n"
        if is_inplace: inner_codes += indent(1) + f"obj.special_from(ref_shape)"
        else: inner_codes += indent(1) + f"obj = Tensor.inherit_from(obj, pivot, shape=...).special_from(ref_shape, allow_view=True)"
    
    # Deal with **kwargs:
    if kwargs_name is None:
        _args += ", sz_func_dim=0, sz_batch_dim=0, sz_feature_dim=0, sz_sequence_dim=0"
        manual_param = with_indent("""
        if sz_func_dim != 0: obj.sz_func_dim = sz_func_dim
        if sz_batch_dim != 0: obj.sz_batch_dim = sz_batch_dim
        if sz_feature_dim != 0: obj.sz_feature_dim = sz_feature_dim
        if sz_sequence_dim != 0: obj.sz_sequence_dim = sz_sequence_dim
        """, 1)
    else:
        manual_param = with_indent(f"""
        if {kwargs_name}.get('sz_func_dim', 0) != 0: obj.sz_func_dim = {kwargs_name}['sz_func_dim']
        if {kwargs_name}.get('sz_batch_dim', 0) != 0: obj.sz_batch_dim = {kwargs_name}['sz_batch_dim']
        if {kwargs_name}.get('sz_feature_dim', 0) != 0: obj.sz_feature_dim = {kwargs_name}['sz_feature_dim']
        if {kwargs_name}.get('sz_sequence_dim', 0) != 0: obj.sz_sequence_dim = {kwargs_name}['sz_sequence_dim']
        """, 1)
    
    func_codes = template(decorators, _def, _fname, _args, doc_in_template, cast, get_size, inner_codes, manual_param)
    resulting_lines = []
    for line in [l.rstrip() for l in func_codes.split('\n')]:
        if line == '' and (len(resulting_lines) == 0 or resulting_lines[-1] == ''): continue
        resulting_lines.append(line)
    func_codes = '\n'.join(resulting_lines).replace('"""', "'''")
    
    # Generate new document string. 
    torch_doc = ''
    if mode=='method': parent = torch.Tensor
    else: parent = torch
    if hasattr(parent, _fname):
        torch_doc = f"The document of the original function is:\n{getattr(parent, _fname).__doc__}"
        for line in torch_doc.split('\n'):
            if line.startswith("See :func:"):
                torch_doc += f"\nwhich is:\n{eval(line.split('See :func:')[-1].strip().strip('`.')).__doc__}"
                break
    new_doc = with_indent(f"""{no_indent(doc_in_template)}

Automatically inheritted method from '{'torch.Tensor' if mode=='method' else 'torch'}.{_fname}'. The automatically generated codes are as follows:
{with_indent(add_lineno(no_indent(func_codes)), 1)}
{with_indent(torch_doc, 1)}
        """, 1)
        
    func_codes = template(decorators, _def, _fname, _args, new_doc, cast, get_size, inner_codes, manual_param)
    resulting_lines = []
    for line in [l.rstrip() for l in func_codes.split('\n')]:
        if line == '' and (len(resulting_lines) == 0 or resulting_lines[-1] == ''): continue
        resulting_lines.append(line)
    return '\n'.join(resulting_lines)

with open(Path(__file__).parent / 'tensor_template.py') as fp:
    __lines__ = fp.read().split('\n')
    sprint = SPrint()
    
def func_section(func_lines, state, ignore_out=False, n_indent=0):
    new_codes = with_indent(get_updated_code(
        no_indent('\n'.join(func_lines)), 
        mode='function' if state == 'global' else 'method', 
        ignore_args=['out'] if ignore_out else []
    ), n_indent)
    if new_codes.strip():
        print(no_indent('\n'.join(func_lines)), end='\n' + '~' * 80 + '\n')
        print(new_codes, end='\n' + '~' * 80 + '\n')
    return new_codes

with scope("main"):
    state = "normal" # In: normal, method, global, biway-method, biway-global
    n_indent = 0
    biway_section = ""
    func_lines = []
    in_func = False
    for line in __lines__:
        c_indent = get_num_indent(line)
        if line.strip() == "### START BIWAY AUTO GENERATION":    n_indent = c_indent; state = 'biway-global'; continue
        elif line.strip() == "### START METHOD AUTO GENERATION": n_indent = c_indent; state = 'method';       continue
        elif line.strip() == "### START GLOBAL AUTO GENERATION": n_indent = c_indent; state = 'global';       continue
        elif line.strip() == "### STOP BIWAY AUTO GENERATION":   n_indent = c_indent; state = 'normal';       continue
        elif line.strip() == "### STOP METHOD AUTO GENERATION":  n_indent = c_indent; state = 'biway-method'; continue
        elif line.strip() == "### STOP GLOBAL AUTO GENERATION":  n_indent = c_indent; state = 'normal';       continue
        
        if state == 'biway-method':
            if len(func_lines) > 0:
                sprint(func_section(func_lines, 'method', n_indent=n_indent))
            sprint(with_indent(biway_section, n_indent))
            state = 'normal'
            continue
        
        start_dec = c_indent == n_indent and line.lstrip().startswith('@')
        start_func = c_indent == n_indent and line.lstrip().startswith('def ')
        if in_func:
            end_func = state == 'normal' or c_indent <= n_indent and line.strip()
            if not end_func:
                if start_func: in_func = True
                func_lines.append(line); continue
            # Do when end function
            sprint(func_section(func_lines, state.split('-')[-1], n_indent=n_indent))
            if state == 'biway-global':
                biway_section += func_section(func_lines, 'method', ignore_out=True, n_indent=n_indent) + '\n'
            in_func = False
            func_lines = []
        if state == 'normal': sprint(line)
        elif start_dec: func_lines = [line]
        elif start_func: in_func = True; func_lines.append(line)
        else: sprint(line)

with open(Path(__file__).parent / 'tensor.py', 'w') as fp:
    fp.write(sprint.text)




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "File to record time."
)

__all__ = """
    time_this
    Timer
    periodic
    periodic_run
    periodic_call
    run_later
""".split()

import time
from functools import wraps
from threading import Thread, Timer as tTimer

from .exception import Error

def time_this(func):
    """
    A function wrapper of function `func` that outputs the time used to run it. 

    Examples::
        >>> @timethis
        ... def func_to_run(*args):
        ...     # inside codes
        ... 
        >>> func_to_run(*input_args)
        # some outputs
        [func_to_run takes 0.001s]
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        if hasattr(getattr(func, '__wrapped__', func), '__name__'):
            print("[%s takes %lfs]"%(func.__name__, end-start))
        else:
            print("[%s takes %lfs]"%(func.__class__.__name__, end-start))
        return result
    return wrapper

class Timer(object):
    """
    An environment that outputs the time used to run codes within. 

    Examples::
        >>> with Timer("test"):
        ...     # inside codes
        ... 
        # some outputs
        [test takes 0.001s]
    """
    def __init__(self, name='', timing=True, log_on_screen=True):
        if not name: name = 'Unknown'
        self.name = str(name)
        self.nround = 0
        self.timing = timing
        self.log_on_screen = log_on_screen
        self.prevrecord = None
    @property
    def recorded_time(self): return self.prevrecord
    def __enter__(self):
        self.start = time.time()
        self.prevtime = self.start
        return self
    def round(self, name = ''):
        self.nround += 1
        self.end = time.time()
        if self.timing:
            if not name: name = "%s(round%d)"%(self.name, self.nround)
            self.prevrecord = self.end - self.prevtime
            print(f"[{name} takes {self.prevrecord}s]")
        self.prevtime = self.end
    def exit(self): raise Error("Jump")("Jump by calling 'exit'. ")
    def __exit__(self, exc_type, exc_value, traceback):
        if self.timing:
            self.prevrecord = time.time() - self.start
            if self.log_on_screen:
                print(f"[{self.name}{'' if self.nround == 0 else '(all)'} takes {self.prevrecord}s]")
        if exc_type == Error("Jump"): return True

class TimerCtrl(tTimer):
    """
    Creates a Time Handler, designed for function `periodic`. 
    """

    def __init__(self, seconds, function):
        tTimer.__init__(self, seconds, function)
        self.isCanceled = False
        self.seconds = seconds
        self.function = function
        self.funcname = function.__name__
        self.startTime = time.time()

    def cancel(self):
        tTimer.cancel(self)
        self.isCanceled = True

    def is_canceled(self): return self.isCanceled

    def setFunctionName(self, funcname): self.funcname = funcname

    def __str__(self):
        return "%5.3fs to run next "%(self.seconds + self.startTime -
                                      time.time()) + self.funcname

class STOP:
    def __init__(self): self.stopped = False; self.index = 0
    def __bool__(self): return self.stopped
    def set_index(self, index): self.index = index
    def stop(self): self.stopped = True
    def resume(self): self.stopped = False

def periodic(period, maxiter=float('Inf'), wait_before_first_call=False, wait_return=False, use_new_thread=True, process_return=None):
    """
    A function wrapper to repeatedly run the wrapped function `period`.
    The return of function being False means the periodic process should stop.
    
    Args:
        period (int / float): the number of seconds to wait between calls.
        maxiter (int): the number of iterations. 
        wait_before_first_call (bool): If True, the iterations would not start immediately, but after a prior `period` seconds.
        wait_return (bool): If True, the next iteration awaits for `period` seconds after the current is over.
            If False, the next iteration starts right after `period` seconds, unless the current job is unfinished even then
                (A warning message would be sent in this circumstance).
        use_new_thread (bool): Whether to ceate new threads for the task.
        process_return (func): a function accept arguments (timer, return_value) to deal with the values returned from the main function. 
            Only activated when `wait_return = True`. It controls the timer by default. The default function is:
                process_return_default = lambda t, r: t.stop() if r is False else None
            Note: This means the timer does not stop by default when no return is provided. 

        Note: if wait_return is False, function should accept a timer controller timer as the first argument just like 'self'.
        timer: controller with .index storing the iteration, .stop() to stop the periodic process before it is meant to stop, .resume() to continue.

    Examples::
        >>> i = 1
        ... @periodic(1)
        ... def func(timer):
        ...     global i
        ...     print(i, timer.index)
        ...     i += 1
        ...     timer.stop()
        ...     timer.resume()
        ... 
        1
        2
        3
        [Output every 1s, and GO ON...]
    """
    if process_return is None:
        process_return = lambda t, r: t.stop() if r is False else None
    if not wait_return and not use_new_thread: raise TypeError("Not waiting for results needs a new thread to work on the process.")
    def wrap(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            fname = func.__name__ + '_' + str(time.time())
            global counts, stops, dones, waitings
            if 'counts' not in globals(): counts = {}
            if 'stops' not in globals(): stops = {}
            if 'dones' not in globals(): dones = {}
            if 'waitings' not in globals(): waitings = {}
            if fname not in counts: counts[fname] = 0
            if fname not in stops: stops[fname] = STOP()
            if fname not in dones: dones[fname] = True
            if fname not in waitings: waitings[fname] = False
            
            def run_next(fname, process_return, args, kwargs, calling):
                global counts, stops, dones, waitings
                if use_new_thread:
                    timer_ctrl = TimerCtrl(period, calling)
                    timer_ctrl.setFunctionName(fname)
                if counts[fname] == 0:
                    counts[fname] = 1
                    if wait_before_first_call:
                        if use_new_thread: timer_ctrl.start()
                        else:
                            time.sleep(period)
                            calling()
                        return
                if not wait_return:
                    if not dones[fname]:
                        print("Warning: Job not finished in one period, waiting for it to be done.")
                        waitings[fname] = True
                        return
                    dones[fname] = False
                    timer_ctrl.start()
                    stops[fname].set_index(counts[fname])
                    ret = func(stops[fname], *args, **kwargs)
                    dones[fname] = True
                    counts[fname] += 1
                    if counts[fname] >= maxiter: return
                    if waitings[fname]:
                        waitings[fname] = False
                        instant_timer_ctrl = TimerCtrl(0, calling)
                        instant_timer_ctrl.setFunctionName(fname)
                        instant_timer_ctrl.start()
                    return
                else:
                    ret = func(*args, **kwargs)
                    counts[fname] += 1
                    if counts[fname] >= maxiter: return
                    process_return(stops[fname], ret)
                    if use_new_thread: timer_ctrl.start()
                    else:
                        time.sleep(period)
                        calling()
            if use_new_thread:
                class new_thread(Thread):
                    def __init__(self, fname, *args, **kwargs):
                        super().__init__()
                        self.fname = fname
                        self.process_return = process_return
                        self.args = args
                        self.kwargs = kwargs
                        
                    def run(self):
                        global stops
                        if stops[self.fname]: return
                        run_next(self.fname, self.process_return, self.args, self.kwargs, self.run)
                new_thread(fname, *args, **kwargs).start()
            else:
                def recursive():
                    run_next(fname, process_return, args, kwargs, recursive)
                recursive()
        return wrapper
    return wrap

def periodic_run(period, maxiter=float('Inf'), use_new_thread=True):
    return periodic(period, maxiter=maxiter, wait_return=True, use_new_thread=use_new_thread)

def periodic_call(period, maxiter=float('Inf'), use_new_thread=True):
    return periodic(period, maxiter=maxiter, wait_return=False, use_new_thread=use_new_thread)

def run_later(wait, use_new_thread=True):
    return periodic(wait, 1, wait_return=True, wait_before_first_call=True, use_new_thread=use_new_thread)
            




from pycamia import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "pyoverload",
    fileinfo = "Refrom the functions to perform override."
)

__all__ = """
    override
    overload
    params
""".split()

from functools import wraps
from .typehint import *
from .typehint import TypeHintError, get_declaration
from .utils import get_environ_vars, _get_wrapped, raw_function, decorator

_debug = False

def set_debug_mode(m): global _debug; _debug = m

def _wrap_params(f):
    if hasattr(f, '__wrapped__'):
        if '[params]' in f.__name__: return None, f
        else: return f, params(run=False)(_get_wrapped(f))
    return f, params(run=False)(f)

def _collect_declarations(func, collection, place_first=False, error=''):
    f = _get_wrapped(raw_function(func))
    toadd = '\t'.join([get_declaration(f)] + ([error] if error else []))
    if place_first: collection.insert(0, toadd)
    else: collection.append(toadd)
    # if f.__doc__:
    #     lines = f.__doc__.split('\n')
    #     toadd = ''
    #     for l in lines:
    #         if not l.strip(): continue
    #         if l.startswith(_get_func_name(f)) or l[0] == '(': toadd = l
    #         else: break
    #     if toadd:
    #         toadd = '\t'.join([toadd, error])
    #         if place_first: collection.insert(0, toadd)
    #         else: collection.append(toadd)

def _get_func_name(f, change_name = True):
    f = raw_function(f)
    fname = f.__name__.split('[')[0]
    if fname.endswith('__0__') or fname.endswith('__default__'):
        fname = '__'.join(fname.split('__')[:-2])
        if change_name: f.__name__ = fname + '['.join(f.__name__.split('[')[1:])
    return fname

@decorator
def overload(func):
    fname = raw_function(func).__name__.split('[')[0]
    local_vars = get_environ_vars(pivot='overload', offset=2).locals
    rawfname = _get_func_name(func, change_name = False)
    key = f"__overload_{rawfname}__"
    overrider = f"__override_{rawfname}__"
    if key in local_vars:
        local_vars[key] = local_vars[key] + 1
        new_name = f"__{fname}_overload_{local_vars[key]}"
        local_vars[new_name] = local_vars[overrider](func)
    else:
        local_vars[key] = 0
        local_vars[overrider] = override(func)
    exec(f"def {rawfname}(*args, **kwargs): return {overrider}(*args, **kwargs)", local_vars)
    # func.__name__ = '['.join([new_name] + func.__name__.split('[')[1:])
    return local_vars[rawfname]

def _try_imp(f, by, collect, place_first=False):
    args, kwargs = by
    run_func, test_func = _wrap_params(f)
    try:
        ret = test_func(*args, **kwargs)
        if run_func is not None: ret = run_func(*args, **kwargs)
        return ret
    except (TypeHintError, HintTypeError) as e:
        _collect_declarations(f, collect, place_first=place_first, error=str(e))
        if _debug: print(str(e))
        return "__try_imp_failed__"

@decorator
def override(*arg):
    """
    Usage 1:

    """

    if len(arg) == 1: arg = arg[0]
    if not iterable(arg):
        if not Functional(arg): raise SyntaxError("Wrong usage of @override. ")
        class override_wrapper:
            def __init__(self, argf):
                self.mainfunc = argf
                self.func_list = [argf]
                fname = raw_function(argf).__name__.split('[')[0]
                if fname.endswith('__0__') or fname.endswith("__default__"): self.default = 0
                else: self.default = None
                self.must_default = False

            def __call__(self, *args, **kwargs):
                if len(args) > 0:
                    argf = args[0]
                    if not kwargs and len(args) == 1 and Functional(argf):
                        fname = raw_function(argf).__name__.split('[')[0]
                        funcname = _get_func_name(self.mainfunc)
                        if fname == "_" or funcname in fname:
                            if fname.endswith('__0__') or fname.endswith("__default__"):
                                if self.default is not None: raise TypeError("Only one default function is acceptable. ")
                                self.default = len(self.func_list)
                            self.func_list.append(argf); return
                            self.must_default = True
                dec_list = []
                if len(self.func_list) == 1:
                    run_func, test_func = _wrap_params(self.func_list[0])
                    try:
                        ret = test_func(*args, **kwargs)
                        if run_func is not None: ret = run_func(*args, **kwargs)
                        if (
                            Functional(ret) and raw_function(ret).__name__.endswith('[params]') or 
                            Sequence(ret) and len(ret) > 0 and all([Functional(x) and raw_function(x).__name__.endswith('[params]') for x in ret])
                        ):
                            dec_list.clear()
                            if Functional(ret): ret = (ret,)
                            for f in ret:
                                ret = _try_imp(f, by=(args, kwargs), collect=dec_list)
                                if isinstance(ret, str) and ret == "__try_imp_failed__": pass
                                else: return ret
                        else: return ret
                    except TypeHintError as e:
                        _collect_declarations(self.func_list[0], dec_list, error=str(e))
                        if _debug: print(str(e))
                elif len(self.func_list) > 1:
                    if self.must_default and self.default is None: self.default = 0
                    for i, f in list(enumerate(self.func_list)) + ([(-1, self.func_list[self.default])] if self.default is not None else []):
                        if i == self.default: continue
                        ret = _try_imp(f, by=(args, kwargs), collect=dec_list, place_first=i==-1)
                        if isinstance(ret, str) and ret == "__try_imp_failed__": pass
                        else: return ret
                else:
                    for name, value in arg.__globals__.items():
                        name = name.replace('override', '').strip('_')
                        if Functional(value) and (name.startswith(raw_function(self.mainfunc).__name__ + '_') or
                                                name.endswith('_' + raw_function(self.mainfunc).__name__)) and \
                                                name != raw_function(self.mainfunc).__name__:
                            ret = _try_imp(value, by=(args, kwargs), collect=dec_list)
                            if isinstance(ret, str) and ret != "__try_imp_failed__": return ret
                func_name = _get_func_name(_get_wrapped(raw_function(self.func_list[self.default if self.default else 0]))).split("_overload")[0]
                dec_list = [func_name + x[x.index('('):] for x in dec_list]
                raise NameError("No {func}() matches arguments {args}. ".format(
                    func=func_name, 
                    args=', '.join([repr(x) for x in args] + 
                        ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
                ) + "All available usages are:\n{}".format('\n'.join(dec_list)))

        owrapper = override_wrapper(arg)
        @wraps(arg)
        def final_wrapper(*args, **kwargs): return owrapper(*args, **kwargs)
        return final_wrapper
    else:
        functionlist = arg
        @decorator
        def wrap(func):
            def wrapper(*args, **kwargs):
                dec_list = []
                for f in functionlist:
                    if isinstance(f, str): f = eval(f)
                    ret = _try_imp(f, by=(args, kwargs), collect=dec_list)
                    if isinstance(ret, str) and ret != "__try_imp_failed__": return ret
                func_name = _get_func_name(func)
                dec_list = [func_name + x[x.index('('):] for x in dec_list]
                raise NameError("No {func}() matches arguments {args}. ".format(
                    func=func_name, 
                    args=', '.join([repr(x) for x in args] + 
                        ['='.join((str(x[0]), repr(x[1]))) for x in kwargs.items()])
                ) + "All available usages are:\n{}".format('\n'.join(dec_list)))
            return wrapper
        return wrap

if __name__ == "__main__":
    # @params # (float, int, str, +list_[3])
    # def f(x:float, y:int, z=3, *k:list_[3], **b:float): return x, y, z, k, b

    # print(f(1.0, 1.0, 'a', [1, 1, 1], [2, 3, 4], t=3))

    @params#(float, float, str, t=int)
    def f(x:float, y:int, z=3, *, t): return x, y, z, t

    print(f(1.0, 1, 'a', t=3))

    import sys
    @params(+Str, [str, null], [str, null])
    def pyprint(*s, sep=' ', end='\n'):
        if sep is None: sep = ' '
        if end is None: end = '\n'
        sys.stdout.write(sep.join(s) + end)

    print('s', 'l', sep=None)

    isint = lambda x: abs(x - round(x)) < 1e-4
    rint = lambda x: int(round(x))
    from rational import GCD
    class rat:

        @params(int, int)
        def __init__numdenint(self, numerator, denominator):
            self.numerator, self.denominator = numerator, denominator
            self.value = self.numerator / self.denominator
            self.cancelation()

        @params([int, float], [int, float])
        def __init__numdenfloat(self, numerator, denominator):
            if isint(numerator) and isint(denominator):
                self.__init__numdenint(rint(numerator), rint(denominator))
            else: self.__init__float(numerator / denominator)

        @params(str)
        def __init__str(self, string):
            try: arg = [float(x) for x in string.split('/')]
            except Exception: raise SyntaxError("Invalid Format")
            self.__init__numdenfloat(*arg)

        @params(int)
        def __init__int(self, num): self.__init__numdenint(num, 1)

        @params(float)
        def __init__float(self, num):
            if isint(num): self.__init__int(rint(num))
            else:
                self.value = num
                self.__init__numdenint(*rat.nearest(num))

        @override(__init__int, __init__float, __init__str, __init__numdenint, __init__numdenfloat)
        def __init__(): pass

        def tuple(self): return self.numerator, self.denominator

        def cancelation(self):
            d = GCD(*self.tuple())
            self.numerator //= d
            self.denominator //= d
            if self.denominator < 0:
                self.numerator = - self.numerator
                self.denominator = - self.denominator

        def __add__(self, other):
            return rat(self.numerator * other.denominator +
                       self.denominator * other.numerator,
                       self.denominator * other.denominator)

        def __mul__(self, other):
            return rat(self.numerator * other.numerator,
                       self.denominator * other.denominator)

        def __str__(self):
            if self.denominator == 1: return str(self.numerator)
            return str(self.numerator)+'/'+str(self.denominator)

        @staticmethod
        def nearest(num, maxiter=100):
            def iter(x, d):
                if isint(x) or d >= maxiter: return int(round(x)), 1
                niter = iter(1 / (x - int(x)), d+1)
                return int(x) * niter[0] + niter[1], niter[0]
            if num >= 0: return iter(num, 0)
            num = iter(-num, 0)
            return -num[0], num[1]

    print(rat(126/270) + rat(25, 14))
    print(rat(126, 270) * rat(25, 14))




from .manager import info_manager

__info__ = info_manager(
    project = "PyCAMIA",
    package = "<main>",
    author = "Yuncheng Zhou",
    create = "2021-12",
    fileinfo = "System operations including class dealing with path or shell menagements.",
    requires = "tqdm"
)

__all__ = """
    path Path
    path_list PathList
    set_black_list
    get_black_list
    curdir
    pardir
    homedir
    rootdir
    pwd ls
    cp copy
    mv move
    merge
    rename
    is_valid_command
    Shell
    get_volumes
    ByteSize
""".split()

import os, sys, re, math, platform, psutil, gzip
from subprocess import PIPE, Popen
from .workflow import scope
from .pythonop import no_indent
from .exception import avouch, touch
from .listop import arg_tuple, to_tuple
from .decorators import alias
from .functions import identity_function
from .environment import update_locals_by_environ

with __info__:
    from tqdm import tqdm

black_list = [".DS_Store", ".git"]

def get_black_list():
    global black_list
    return black_list

def set_black_list(blacklist):
    global black_list
    black_list = blacklist

@alias("path_list")
class PathList(list):
    def __new__(cls, *args, ref=None):
        args = arg_tuple(args)
        if ref is None: self = super().__new__(cls, args)
        else: self = super().__new__(cls, (Path(a, ref=ref) for a in args))
        self._ref_dir = ref
        return self
    
    def __getitem__(self, key):
        if touch(lambda: len(key), default=-1) == len(self): return PathList([x for x, b in zip(self, key) if b])
        res = super().__getitem__(key)
        if isinstance(res, (tuple, list)): return PathList(res)
        return res
        
    def sort(self, *, key=identity_function):
        super().sort(key=key)
        
    def append(self, p):
        if self._ref_dir == p._ref: super().append(p)
        super().append(Path(p, ref=self._ref_dir))

    @alias("__or__", "filter")
    def select_file_type(self, k): return self[[x | k for x in self]]
    @alias("__sub__")
    def relative_to(self, y): return PathList([x - y for x in self])
    def __mod__(self, k): return PathList([x % k for x in self])
    def __truediv__(self, k): return PathList([x / k for x in self])
    def __rtruediv__(self, k): return PathList([k / x for x in self])

@alias("path")
class Path(str):
    """
    Path("abc") -> path_object

    An object managing file path.
    Use keyword argument `read_only=True` to make sure that one cannot accidentally delete or modify the file/directory using `Path`.

    Examples::
        >>> Path()
        {working directory}
        >>> Path("abc", "sub_folder")
        abc/sub_folder
        >>> Path("abc")/"sub_folder"
        abc/sub_folder
        >>> Path("abc", ref="bcd") # abc and bcd are brother directories
        ../abc
    """
    
    sep = os.path.sep #[In Mac]: '/'
    extsep = os.path.extsep #[In Mac]: '.'
    pathsep = os.path.pathsep #[In Mac]: ':'
    _curdir = os.path.curdir #[In Mac]: '.'
    _pardir = os.path.pardir #[In Mac]: '..'
    _homedir = os.path.expanduser("~") # [In Mac]: '/Users/admin'; [In Win]: Unknown
    _rootdir = os.path.abspath(os.path.curdir).split(sep)[0] + sep
        # [In Mac or Linux]: '/'; [In Win]: 'C:\\' for disk C
    namesep = '_'
    File = b'\x04'
    Folder = Dir = b'\x07'
    invalid_char_regx = r"[:\?$]"

    @alias('list_all', 'listall', 'ls_a', all_items=True, depth=1)
    @alias('list_dir', 'listdir', 'list', 'ls', depth=1)
    @alias('list_subdirs', 'listsubdirs', depth=-1)
    @alias('list_files', 'listfiles', depth=0)
    @alias('walk')
    def _create_dir_list(self, all_items=False, depth=None):
        res = PathList(self._recursively_listdir(all_items, depth))
        res.sort()
        return res

    @alias('iter_all', 'iterall', all_items=True, depth=1)
    @alias('iter_dir', 'iterdir', 'iter', depth=1)
    @alias('iter_subdirs', 'itersubdirs', depth=-1)
    @alias('iter_files', 'iterfiles', depth=0)
    @alias('iter_walk')
    def _recursively_listdir(self, all_items=False, depth=None):
        """
        Args:
            all_items (bool): whether to search hidden files / directories or not
            depth (int, flag like):
                   [-1] means folders with no subfolders
                    [0] means all files in the directory (recursively)
                    [d] means paths with relative depth d (d > 0)
                 [None] means all relative paths in the folder in a recursive scan
            listing with depth = 1 is equivalent to os.listdir. 
        
        Contribution:
            Function code provided by @Yiteng Zhang
        """
        recursively_searched = False
        for f in os.listdir(self._abs):
            if f in get_black_list():
                continue
            p = Path(f, ref=self, read_only=self._read_only)
            if not all_items and p.is_hidden():
                continue
            if depth is None:
                yield p
                if p.is_dir():
                    for cp in p._recursively_listdir(all_items=all_items, depth=depth):
                        yield p/cp
            else:
                assert isinstance(depth, int)
                if p.is_file() and depth >= 0:
                    yield p
                elif p.is_dir():
                    if depth != 1:
                        for cp in p._recursively_listdir(all_items=all_items, depth=depth-1 if depth > 0 else depth):
                            if cp == p: yield p
                            else: yield p/cp
                        recursively_searched = True
                    else:
                        yield p
        if depth == -1 and not recursively_searched:
            yield self

    def __new__(cls, *init_texts, ref=None, read_only=False):
        """
        path object containing path in `init_texts` and a reference folder `ref`. 

        Examples::
            >>> Path("a", "b")
            a/b
            >>> Path("a/b/c/d", ref="a")
            b/c/d
        """
        if ref is not None:
            if not isinstance(ref, str): raise TypeError(f"Path reference should be a string, not {ref} of type {type(ref)}. ")
            if isinstance(ref, Path): ref = ref._abs
            else:
                if ref != Path._rootdir: ref = os.path.normpath(str(ref))
                if not os.path.isabs(ref): ref = os.path.abspath(ref)
        if len(init_texts) == 1 and isinstance(init_texts[0], (list, tuple)):
            init_texts = init_texts[0]
        if len(init_texts) == 1 and isinstance(init_texts[0], Path):
            path_object = init_texts[0]
            if ref is None: return path_object
            else:
                abs_path = path_object._abs
                self = super().__new__(cls, abs_path if ref == Path._rootdir else os.path.relpath(abs_path, ref))
                self._ref_dir = ref
                self._read_only = read_only
            return self
        init_texts = [str(x) for x in init_texts]
        # avouch(all([isinstance(x, str) for x in init_texts]), f"Cannot create path from {init_texts} as there are non-string elements. ")
        if len(init_texts) <= 0 or (len(init_texts) == 1 and init_texts[0] in (os.path.curdir, '')): string = os.path.curdir
        else:
            init_texts = [str(x).strip() for x in init_texts]
            for x in init_texts:
                if len(re.findall(Path.invalid_char_regx, x)) > 0:
                    print(f"Warning: Invalid characters in path '{x}'.")
            if os.name in ('posix',) and init_texts[0].startswith('~'):
                init_texts = (Path._homedir + init_texts[0][1:],) + init_texts[1:]
            string = os.path.normpath(os.path.join(*init_texts))
            if init_texts[0] == '': string = os.path.sep + string
        if ref == Path._rootdir:
            self = super().__new__(cls, os.path.abspath(string) if Path._curdir in string else string)
            self._ref_dir = Path._rootdir
            self._read_only = read_only
            return self
        if os.path.isabs(string):
            if ref is None or ref == Path._rootdir:
                self = super().__new__(cls, string)
                self._ref_dir = Path._rootdir
            else:
                self = super().__new__(cls, os.path.relpath(string, ref))
                self._ref_dir = ref
            self._read_only = read_only
            return self
        if ref is None: ref = os.path.abspath(os.path.curdir)
        self = super().__new__(cls, string)
        self._ref_dir = ref
        self._read_only = read_only
        return self

    @property
    def _ref(self): return self._ref_dir
    @property
    def _abs(self): return os.path.normpath(os.path.join(self._ref_dir, str(self)))
    @property
    def _rel(self): return os.path.relpath(self, Path._rootdir) if self._ref_dir == Path._rootdir else str(self)

    @alias('ref')
    @property
    def reference_dir(self): return Path(self._ref, ref=Path._rootdir, read_only=self._read_only)
    @alias("__invert__", "__abs__", "abs")
    @property
    def absolute_path(self): return Path(self._abs, ref=Path._rootdir, read_only=self._read_only)
    @alias('rel', "relative")
    @property
    def relative_path(self): return Path(self._rel, ref=self._ref, read_only=self._read_only)
    @property
    def relative_to_cur(self): return self - curdir

    def __mod__(x, y): return Path(super().__mod__(to_tuple(y)), ref=x._ref, read_only=x._read_only)
    
    @alias("__sub__")
    def relative_to(x, y):
        if isinstance(y, Path): y = y._abs
        return Path(x._abs, ref=y, read_only=x._read_only)
    @alias("__add__")
    def name_add(x, y):
        y = str(y)
        if x.is_filepath():
            return x.parent / (x.name + y + Path.extsep + x.ext)
        else: return Path(super(Path, x).__add__(y), ref=x._ref, read_only=x._read_only)
    @alias("__xor__")
    def name_subscript(x, y):
        y = str(y).lstrip(Path.namesep)
        if x.is_filepath():
            return x.parent / (x.name.rstrip(Path.namesep) + Path.namesep + y + Path.extsep + x.ext)
        else: return Path(super(Path, x).__add__(Path.namesep + y), ref=x._ref, read_only=x._read_only)
    @alias("__pow__")
    def common_parent(x, y):
        return Path(os.path.commonpath(x._abs, y._abs), ref=x._ref)

    @alias("__floordiv__")
    def add_ext(x, y):
        if not y: return x
        return Path(Path.extsep.join((x.rstrip(Path.extsep), y.lstrip(Path.extsep))), ref=x._ref, read_only=x._read_only)
    @alias("__truediv__", "cd", "chdir", "concat", "append")
    def add_dir(x, y):
        if isinstance(y, Path): avouch(y._ref is None or y._ref == x._abs, "Only folder paths with reference to parent path can be appended. ")
        return Path(x, y, ref=x._ref, read_only=x._read_only)
    @alias("__rtruediv__", "prefix", "with_pardir")
    def pre_dir(x, y):
        if isinstance(y, Path):
            ref = y._ref
        else:
            if x._ref.endswith(y): ref = x._ref[:-len(y)]
            else: ref = x._ref
        return Path(y, x, ref=ref, read_only=x._read_only)
    @alias("__or__")
    def is_file_type(x, y):
        for y in to_tuple(y):
            if (y == "FILE" or y == Path.File) and x.is_file(): return True
            if (y == "FOLDER" or y == Path.Folder) and x.is_dir(): return True
            if x.ext.lower() == y.lower().lstrip('.'): return True
        return False
    def __eq__(x, y): return x._abs == y._abs if isinstance(y, Path) else super().__eq__(y)
    def __len__(self): return len(str(self))
    def __hash__(self): return super().__hash__()
    @property
    def repr(self): return self._ref + os.path.sep + super().__repr__()

    def __iter__(self):
        for p in self.list(): yield p

    def files(self):
        for p in self.list() | 'FILE': yield p

    def folders(self):
        for p in self.list() | 'FOLDER': yield p

    def __contains__(self, x):
        for p in self:
            if p == x: return True
        return False

    def str_contains(self, x): return x in str(self)
    
    @classmethod
    def join(cls, *args):
        args = arg_tuple(args)
        return cls(*args)

    @property
    def ext(self):
        if hasattr(self, '_ext') and self._ext is not None:
            return self._ext
        if self.is_dir():
            self._ext = ""
            return ""
        file_name = self.fullname
        parts = file_name.split(Path.extsep)
        if parts[-1].lower() in ('zip', 'gz', 'rar') and len(parts) > 2: brk = -2
        elif len(parts) > 1: brk = -1
        else: brk = 1
        self._ext = Path.extsep.join(parts[brk:])
        return self._ext

    @property
    def name(self):
        if hasattr(self, '_name') and self._name is not None:
            return self._name
        file_name = self.fullname
        if self.is_dir():
            self._name = file_name
            return file_name
        parts = file_name.split(Path.extsep)
        if parts[-1].lower() in ('zip', 'gz', 'rar') and len(parts) > 2: brk = -2
        elif len(parts) > 1: brk = -1
        else: brk = 1
        self._name = Path.extsep.join(parts[:brk])
        return self._name

    def with_name(self, name):
        return self.parent / name // self.ext

    def with_ext(self, ext: str):
        return self.parent / self.name // ext

    @alias("fullname", "basename")
    @property
    def filename(self):
        if not hasattr(self, '_filename'):
            self._filename = self.split()[-1]
        return self._filename
    
    @property
    def dirname(self):
        if not hasattr(self, '_dirname'):
            path_tokens = self.split()
            if len(path_tokens) == 0 or path_tokens[-1] in (Path._curdir, Path._pardir, Path._homedir, Path.rootdir):
                path_tokens = self.absolute_path.split()
            self._dirname = Path.sep.join(path_tokens[:-1])
        return self._dirname
    
    @property
    def parent(self):
        return Path(self.dirname, ref=self._ref, read_only=self._read_only)
    
    @property
    def children(self):
        return self.list()

    def is_hidden(self):
        if os.name== 'nt':
            try: import win32api, win32con
            except ModuleNotFoundError: raise ModuleNotFoundError("Packages win32api, win32con needed for hidden file recognition in windows system. Please manually install them. ")
            attribute = win32api.GetFileAttributes(self._abs)
            return attribute & (win32con.FILE_ATTRIBUTE_HIDDEN | win32con.FILE_ATTRIBUTE_SYSTEM)
        elif os.name == 'posix':
            return self.name.startswith('.') #linux or macosx
        else: raise TypeError(f"Unrecognized operating system '{os.name}' for hidden state.")

    def split(self, *args, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            return super().split(Path.sep)
        return super().split(*args, **kwargs)

    def is_abs(self): return self._ref == Path._rootdir
    def is_rel(self): return self._ref != Path._rootdir
    def exists(self): return os.path.exists(self._abs)
    @alias("isfile", "is_folder")
    def is_file(self): return os.path.isfile(self._abs)
    @alias("isdir", "is_folder")
    def is_dir(self): return os.path.isdir(self._abs)
    @alias("isempty")
    def is_empty(self): return len(self.list()) == 0
    @alias("is_file_path")
    def is_filepath(self): return True if os.path.isfile(self._abs) else all([0 < len(e) <= 4 for e in self.ext.split(Path.extsep)])
    @alias("is_dir_path", "is_folderpath", "is_folder_path")
    def is_dirpath(self): return True if os.path.isdir(self._abs) else not all([0 < len(e) <= 4 for e in self.ext.split(Path.extsep)])
    
    @alias("delete", "del", "rm")
    def remove(self, verbose=True):
        if self._read_only: raise RuntimeError(f"Trying to remove read only Path object '{self}'")
        if self.is_dir():
            if verbose and self.ls():
                print(f"You want to delete directory: {self}")
                if 'y' not in input("Do you want to continue? [Y/n]: ").lower(): return
            for f in self: f.remove(verbose=verbose)
            os.rmdir(self._abs)
        else: os.remove(self._abs)
        
    def gzip(self, force=False, replace=False):
        path_zip = self // 'gz'
        if not force and path_zip.exists(): 
            print(f"Path already exists: {path_zip}")
            if 'y' not in input("Do you want to overwrite? [Y/n]: ").lower(): return
        with self.open('rb') as file_in, gzip.open(path_zip, 'wb') as file_zip:
            file_zip.writelines(file_in)
        if replace: self.remove()
        
    def rename(self, new_name):
        if self._read_only: raise RuntimeError(f"Trying to rename read only Path object '{self}'")
        old_wd = os.path.abspath(os.curdir)
        os.chdir(self.parent._abs)
        os.rename(self.filename, (new_name - self.parent) if isinstance(new_name, Path) else new_name)
        os.chdir(old_wd)

    @alias('cmd', "system")
    def command(self, command):
        try:
            cmd = command.format(self, file=self, path=self)
            old_wd = os.path.abspath(os.curdir)
            os.chdir(self.ref)
            os.system(cmd)
            os.chdir(old_wd)
        except Exception as e:
            print(f"Command error in {cmd}:", e)
            
    def move_to(self, path, **kwargs):
        move(self, path, **kwargs)

    def copy_to(self, path, **kwargs):
        copy(self, path, **kwargs)

    def open(self, mode='r+'):
        if self._read_only and not mode.startswith('r'): raise RuntimeError(f"Trying to open read only Path object '{self}' in edit mode {mode}.")
        if 'w' not in mode and not self.exists(): raise FileNotFoundError(f"Cannot find file {self._abs}. ")
        elif not self.parent.exists(): raise FileNotFoundError(f"Cannot find file folder {self.parent._abs}. ")
        avouch(self.is_file() or self.is_filepath() and 'w' in mode, "Only files can be opened as python stream. ")
        return open(self._abs, mode)

    @alias('browse')
    def open_in_browser(self):
        if os.name == 'posix':
            self.command("open {path}")
        else:
            raise TypeError(f"Unrecognized operating system '{os.name}' for default open method 'open_in_brower' (or 'browse').")

    def mkdir(self, new_folder = None):
        """
        Make directory along the path. Create `new_folder` if it is provided. 

        i.e., Path("/Users/username/code/dataset").mkdir()
        will recursive check if "/Users", "/Users/username", "/Users/username/code", "/Users/username/code", "/Users/username/code/dataset"
        is exists or not and make the corresponding directory.
        """
        if self._read_only: raise RuntimeError(f"Trying to make directory in read only Path object '{self}'")
        if not self.is_dir():
            p = self.ref
            if not p.is_dir(): p.mkdir()
            for seg in self.split():
                p = p / seg
                # if not p.exists() and p.is_filepath():break
                if not p.is_dir(): os.mkdir(p)
        if not new_folder: return self
        os.mkdir(self / new_folder)
        return self / new_folder

    def size(self):
        return ByteSize(self._scan_size())
    
    def _scan_size(self):
        fsize = os.stat(self._abs).st_size
        if not self.is_dir(): return fsize
        for p in self.list_all():
            fsize += p._scan_size()
        return fsize
    
    @classmethod
    @property
    def curdir(self): return Path(Path._curdir)
    @classmethod
    @property
    def pardir(self): return Path(Path._pardir)
    @classmethod
    @property
    def homedir(self): return Path(Path._homedir)
    @classmethod
    @property
    def rootdir(self): return Path(Path._rootdir)

curdir = Path(Path._curdir)
pardir = Path(Path._pardir)
homedir = Path(Path._homedir)
rootdir = Path(Path._rootdir)

def pwd(): return curdir.abs
def ls(): return curdir.ls()

def merge(src, dst, copy_unit = None, force = None, _pbar = None):
    avouch(src.is_dir() and dst.is_dir(), TypeError(f"Only two directories can be merged, not {src} and {dst}."))
    avouch(isinstance(force, str) and force in ('skip', 'replace', 'keep', 'merge', 'skip_merge', 'replace_merge', 'keep_merge', 'auto') or force is None, 
           TypeError(f"Keyword argument 'force' for copy should be a string from (skip, replace, keep, merge, skip_merge, replace_merge, keep_merge, auto), not {force!r}, see docstring for more information. "))
    copy(src, dst, copy_unit=copy_unit, force='merge' if force is None else (force + '_merge' if not force.endswith('_merge') else force), _pbar=_pbar)

@alias('cp')
def copy(src, dst, copy_unit = None, force = None, show_progress = False, _pbar = None, _fp = None):
    """
    Copy file from 'src' to 'dst'.

    Args:
        src (str or Path): The source file path.
        dst (str or Path): The target file path.
        copy_unit (int, optional): Number of Bytes copied each time. Defaults to 1MB: 1<<(10 * 2) for small files and 32MB: 1<<25 for files bigger than 16GB.
        force (str, optional): How to deal with conflicts when file exists in the target directory. 
            [force='rename' or 'rename_replace']: force the argument `dst` be the target folder (instead of the parent of target folder), even if it exists (this will replace the folder).
            [force='rename_skip']: force the folder to be renamed as `dst`, skip if it already exists.
            [force='rename_keep']: force the folder to be renamed as `dst`, keep the two folders if it already exists.
            [force='skip']: perform a force skip of file if it exists.
            [force='replace']: perform a forced replacement of the file.
            [force='keep']: perform a force keep of two files by creating a new duplication. 
            [force='merge' or 'skip_merge']: perform a force merge of the folders, it is equivalent to 'skip' for a single file, 
                and merges two folders by skipping existing files and copying additional ones. 
            [force='replace_merge']: perform a force merge of the folders, it is equivalent to 'replace' for a single file, 
                and merges two folders by replacing all existing files and copying additional ones, this operation will not remove files only in the destiny directory. 
            [force='keep_merge']: perform a force merge of the folders, it is equivalent to 'keep' for a single file, 
                and merges two folders by keeping the both files from source and target folder, with the new copied files (from source directory) maked with '({number})'. 
            [force='auto']: perform keep/ keep_merge for copies from a directory to itself but replace/ replace_merge for other senarios.
            [force=None]: raise Error for conflicts.
            Defaults to None. 
            Note: Do use '_merge' forces for the circumstances when the destiny directory exists or the source directory will be copied into the destiny folder.
        _pbar (tqdm progress bar, !PRIVATE!): The progress bar for recursive copy of folder, please do not use it without
            deep understanding of the inner logic (which may cause error, or more severely loss of data). Defaults to None.
        _fp (stdio for recording file, !PRIVATE!): The file object for temporary records, please do not use it without
            deep understanding of the inner logic (which may cause error, or more severely loss of data). Defaults to None.
    """
    if _pbar is None:
        avouch(isinstance(src, str))
        avouch(isinstance(dst, str))
        src = Path(src, read_only=True)
        dst = Path(dst)
        if not src.exists(): raise FileNotFoundError(f"Cannot find the copy source '{src}'.")
        if dst.is_file():
            if force is None: raise TypeError(f"File already exists: {dst}.")
            if force not in ('skip', 'replace', 'keep'):
                raise TypeError("'rename' and 'merge' values for argument 'force' in copy are only for directory sources")
        if dst.is_dir() and force in ('skip', 'replace', 'keep'):
            dst = dst / src.filename
        elif force == 'rename': force = 'replace'
        elif isinstance(force, str) and force.startswith('rename_'): force = force[len('rename_'):]
        if dst.exists():
            if force is None: raise FileExistsError(f"Target '{dst}' already exists in destiny directory. ")
            if not isinstance(force, str) or force not in ('skip', 'replace', 'keep', 'merge', 'skip_merge', 'replace_merge', 'keep_merge', 'auto'):
                raise TypeError(f"Keyword argument 'force' for copy should be a string from (skip, replace, keep, rename, rename_skip, rename_replace, rename_keep, merge, skip_merge, replace_merge, keep_merge, auto), not {force!r}, see docstring for more information. ")
            if force == 'auto':
                if dst.parent == src.parent: force = 'keep'
                else: force = 'replace'
                if src.is_dir(): force += '_merge'
            if force == 'merge': force = 'skip_merge'
        size_all = src.size()
        if show_progress:
            _pbar = tqdm(total=size_all, desc="Copy", unit="item", unit_scale=True, 
                        ncols=80, bar_format="{l_bar}{bar}|")
        else: _pbar = 1
    copy_record = dst.parent/("." + dst.filename)//'copyrcd'
    if dst.exists() and not copy_record.exists():
        if force == 'skip':
            if _pbar != 1: _pbar.update(src.size())
            return
        if force == 'replace' and not copy_record.exists(): dst.remove(verbose=False)
        elif force == 'keep':
            i_copy = 1
            while True:
                i_copy += 1
                new_name = dst.name + f"({i_copy})"
                if dst.with_name(new_name).exists() and not (dst.parent/("." + new_name)//dst.ext//'copyrcd').exists(): continue
                break
            dst = dst.with_name(new_name)
    if src.is_dir():
        if not dst.exists(): dst.mkdir()
        file_force = force
        if isinstance(file_force, str) and file_force.endswith('_merge'): file_force = file_force[:-len('_merge')]
        if _fp is None:
            main_copy_record = dst.parent/("." + dst.filename)//'copyrcd'
            if main_copy_record.exists():
                with open(main_copy_record, 'r') as fp:
                    src_path, *prev_done = [l.strip() for l in fp.readlines() if l]
                    if len(prev_done) > 0: prev_done = prev_done[0]
                    else: prev_done = None
                if src._abs != src_path:
                    main_copy_record.remove(verbose=False)
                    started = True
                else: started = False
            else: started = True
            if started:
                fp = open(main_copy_record, "w").__enter__()
                fp.write(src._abs + '\n')
            else: fp = open(main_copy_record, "a").__enter__()
        else: fp = _fp
        for f in src.iter_files():
            avouch(f.is_file(), RuntimeError(f"Got non-file {f} in iteration of {src}"))
            if not started:
                if prev_done is not None and f._abs != prev_done: continue
                started = True
            (dst/f.filename).parent.mkdir()
            copy(src/f, dst/f.filename, copy_unit=copy_unit, force=file_force, _pbar=_pbar, _fp=fp)
        if _fp is None: fp.__exit__(); main_copy_record.remove(verbose=False)
    else:
        file_size = src.size()
        if copy_unit is None:
            if file_size >= 1 << (10 * 3 + 4): # >= 16GB
                copy_unit = 1 << 25 # 32MB
            else: copy_unit = 1 << 20 # 1MB
        prev_done = 0
        if copy_record.exists():
            with open(copy_record, 'r') as fp:
                src_path, *prev_done = [l.strip() for l in fp.readlines() if l]
                if len(prev_done) > 0: prev_done = eval(prev_done[0])
                else: prev_done = 0
            if src._abs != src_path: copy_record.remove(); prev_done = 0
        if prev_done == 0:
            with open(copy_record, 'w') as fp:
                fp.write(src._abs + '\n')
            with open(dst, 'w'): ...
        n_file_share = (file_size - prev_done) / copy_unit
        n_remain = file_size - prev_done - int(n_file_share) * copy_unit
        n_file_share = math.ceil(n_file_share)
        with open(src, 'rb') as fp_src:
            fp_src.seek(prev_done)
            for i in range(n_file_share):
                # if i >= 10: raise KeyboardInterrupt()
                if i == n_file_share - 1: n_copy = n_remain
                else: n_copy = copy_unit
                with open(dst, 'rb+') as fp_dst:
                    fp_dst.seek(prev_done + i * copy_unit)
                    fp_dst.write(fp_src.read(n_copy))
                with open(copy_record, 'a') as fp:
                    fp.write(f"{prev_done + i * copy_unit + n_copy}\n")
                if _pbar != 1: _pbar.update(n_copy)
        copy_record.remove()
        if _fp is not None: _fp.write(src._abs + '\n')

@alias('mv')
def move(src, dst, copy_unit = None, force=None, show_progress = False, _pbar = None, _fp = None):
    """
    Move file from 'src' to 'dst'.

    Args:
        src (str or Path): The source file path.
        dst (str or Path): The target file path.
        copy_unit (int, optional): Number of Bytes copied each time. Defaults to 1MB: 1<<(10 * 2) for small files and 32MB: 1<<25 for files bigger than 16GB.
        force (str, optional): How to deal with conflicts when file exists in the target directory. 
            [force='rename' or 'rename_replace']: force the argument `dst` be the target folder (instead of the parent of target folder), even if it exists (this will replace the folder).
            [force='rename_skip']: force the folder to be renamed as `dst`, skip if it already exists.
            [force='rename_keep']: force the folder to be renamed as `dst`, keep the two folders if it already exists.
            [force='skip']: perform a force skip of file if it exists.
            [force='replace']: perform a forced replacement of the file.
            [force='keep']: perform a force keep of two files by creating a new duplication. 
            [force='merge' or 'skip_merge']: perform a force merge of the folders, it is equivalent to 'skip' for a single file, 
                and merges two folders by skipping existing files and moving additional ones. 
            [force='replace_merge']: perform a force merge of the folders, it is equivalent to 'replace' for a single file, 
                and merges two folders by replacing all existing files and moving additional ones, this operation will not remove files only in the destiny directory. 
            [force='keep_merge']: perform a force merge of the folders, it is equivalent to 'keep' for a single file, 
                and merges two folders by keeping the both files from source and target folder, with the new moved files (from source directory) maked with '({number})'. 
            [force='auto']: perform replace/ replace_merge for movings.
            [force=None]: raise Error for conflicts.
            Defaults to None. 
            Note: Do use '_merge' forces for the circumstances when the destiny directory exists or the source directory will be moved into the destiny folder.
        _pbar (tqdm progress bar, !PRIVATE!): The progress bar for recursive moving of folder, please do not use it without 
            deep understanding of the inner logic (which may cause error, or more severely loss of data). Defaults to None.
        _fp (stdio for recording file, !PRIVATE!): The file object for temporary records, please do not use it without
            deep understanding of the inner logic (which may cause error, or more severely loss of data). Defaults to None.
    """
    if _pbar is None:
        avouch(isinstance(src, str))
        avouch(isinstance(dst, str))
        src = Path(src)
        dst = Path(dst)
        if not src.exists(): raise FileNotFoundError(f"Cannot find the moving source '{src}'.")
        if dst.is_file() and force not in ('skip', 'replace', 'keep'):
            raise TypeError("'rename' and 'merge' values for argument 'force' in move are only for directory sources")
        if dst.is_dir() and force in ('skip', 'replace', 'keep'):
            dst = dst / src.filename
        elif force == 'rename': force = 'replace'
        elif force.startswith('rename_') and force.startswith('rename_'): force = force[len('rename_'):]
        if dst.exists():
            if force is None: raise FileExistsError(f"Target '{dst}' already exists in destiny directory. ")
            if not isinstance(force, str) or force not in ('skip', 'replace', 'keep', 'merge', 'skip_merge', 'replace_merge', 'keep_merge', 'auto'):
                raise TypeError(f"Keyword argument 'force' for copy should be a string from (skip, replace, keep, rename, rename_skip, rename_replace, rename_keep, merge, skip_merge, replace_merge, keep_merge, auto), not {force!r}, see docstring for more information. ")
            if force == 'auto':
                force = 'replace'
                if src.is_dir(): force += '_merge'
            if force == 'merge': force = 'skip_merge'
        size_all = src.size()
        if show_progress:
            _pbar = tqdm(total=size_all, desc="Move", unit="item", unit_scale=True, 
                        ncols=80, bar_format="{l_bar}{bar}|")
        else: _pbar = 1
    move_record = dst.parent/("." + dst.filename)//'movercd'
    if dst.exists() and not move_record.exists():
        if force == 'skip':
            if _pbar != 1: _pbar.update(src.size())
            return
        if force == 'replace' and not move_record.exists(): dst.remove(verbose=False)
        elif force == 'keep':
            i_move = 1
            while True:
                i_move += 1
                new_name = dst.name + f"({i_move})"
                if dst.with_name(new_name).exists() and not (dst.parent/("." + new_name)//dst.ext//'movercd').exists(): continue
                break
            dst = dst.with_name(new_name)
    if src.is_dir():
        if not dst.exists(): dst.mkdir()
        file_force = force
        if isinstance(file_force, str) and file_force.endswith('_merge'): file_force = file_force[:-len('_merge')]
        if _fp is None:
            main_move_record = dst.parent/("." + dst.filename)//'movercd'
            if main_move_record.exists():
                with open(main_move_record, 'r') as fp:
                    src_path, *_, prev_done = [l.strip() for l in fp.readlines() if l]
                if src._abs != src_path:
                    main_move_record.remove(verbose=False)
                    started = True
                else: started = False
            else: started = True
            if started:
                fp = open(main_move_record, "w").__enter__()
                fp.write(src._abs + '\n')
            else: fp = open(main_move_record, "a").__enter__()
        else: fp = _fp
        for f in src.iter_files():
            avouch(f.is_file(), RuntimeError(f"Got non-file {f} in iteration of {src}"))
            if not started:
                if f._abs != prev_done: continue
                started = True
            (dst/f).parent.mkdir()
            move(src/f, dst/f, copy_unit=copy_unit, force=file_force, _pbar=_pbar, _fp=fp)
        if _fp is None: fp.__exit__(); main_copy_record.remove(verbose=False)
    else:
        file_size = src.size()
        if dst.parent == src.parent:
            src.rename(dst.filename)
            if _pbar != 1: _pbar.update(file_size)
            return
        if copy_unit is None:
            if file_size >= 1 << (10 * 3 + 4): # >= 16GB
                copy_unit = 1 << 25 # 32MB
            else: copy_unit = 1 << 20 # 1MB
        prev_done = 0
        if move_record.exists():
            with open(move_record, 'r') as fp:
                src_path, *_, prev_done = [l.strip() for l in fp.readlines() if l]
            if src._abs != src_path: move_record.remove(); prev_done = 0
            else: prev_done = eval(prev_done)
        if prev_done == 0:
            with open(move_record, 'w') as fp:
                fp.write(src._abs + '\n')
            with open(dst, 'w'): ...
        n_file_share = (file_size - prev_done) / copy_unit
        n_remain = file_size - prev_done - int(n_file_share) * copy_unit
        n_file_share = math.ceil(n_file_share)
        with open(src, 'rb') as fp_src:
            fp_src.seek(prev_done)
            for i in range(n_file_share):
                # if i >= 10: raise KeyboardInterrupt()
                if i == n_file_share - 1: n_copy = n_remain
                else: n_copy = copy_unit
                with open(dst, 'rb+') as fp_dst:
                    fp_dst.seek(prev_done + i * copy_unit)
                    fp_dst.write(fp_src.read(n_copy))
                with open(copy_record, 'a') as fp:
                    fp.write(f"{prev_done + i * copy_unit + n_copy}\n")
                if _pbar != 1: _pbar.update(n_copy)
        move_record.remove()
        if _fp is not None: _fp.write(src._abs + '\n')
        src._read_only = False
        src.remove()
    
def rename(src, dst):
    avouch(isinstance(src, str))
    avouch(isinstance(dst, str))
    src = Path(src)
    dst = Path(dst)
    src = Path(src, ref=src.parent)
    if dst.parent == curdir:
        dst = dst.filename
    else: dst = dst - src.parent
    src.command(f"mv {{path}} {dst}")
    
def is_valid_command(cmd, error_name='command not found'):
    p = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
    try: stdout, stderr = p.communicate(timeout=0.05)
    except: stderr = b''
    return bytes(error_name, 'utf8') not in stderr
    
def get_volumes(**kwargs):
    drives = []
    for p in psutil.disk_partitions():
        for k, v in kwargs.items():
            this_v = getattr(p, k, None)
            if isinstance(this_v, str):
                this_v = this_v.split(',')
            else: this_v = [this_v]
            if callable(v):
                if not v(this_v): break
            elif isinstance(v, str) and v.startswith('!'):
                v = v[1:]
                if len(this_v) == 1 and v == str(this_v[0]): break
                elif v in this_v: break
            elif v not in this_v: break
        else:
            drives.append(p)
    return drives
    
class Shell:

    def __init__(self, *directories, verbose=False):
        self.print = print if verbose else lambda x: None
        self.tools = []
        directories = arg_tuple(directories)
        for d in directories:
            for f in Path(d).files():
                if f | 'py': setattr(self, f.name, f"python3 {f.abs}")
                elif f | ('sh', 'exe', ''): setattr(self, f.name, f.abs)
                else: continue
                self.tools.append(f.name)

    def __getattr__(self, key):
        update_locals_by_environ()
        def run(*args):
            self(getattr(self, key, key) + ' ' + ' '.join(args))
        return run

    def __call__(self, string):
        update_locals_by_environ()
        vars = re.findall(r'{(\w+)}', string)
        command = string.format(**{v: eval(v).replace(' ', '\ ') if isinstance(eval(v), str) else eval(v) for v in vars})
        key, *args = re.sub(r'[^\\] ', lambda x: x.group().replace(' ', '\n'), command).split('\n')
        do = getattr(self, key, key)
        command = do + command[len(key):]
        self.print("Running:", command)
        opt = None
        for arg in enumerate(args):
            arg = arg.strip()
            if arg.startswith('-'): opt = arg; continue
            if os.path.sep in arg:
                arg = arg.replace('\ ', ' ')
                arg = arg.strip("""'"'""")
                p = Path(arg)
                file_out = False
                if opt is not None and (opt == '-o' or opt.startswith('--out')):
                    file_out = True
                    p = p.parent
                if not p.exists():
                    self.print(f"Warning: Path doesn't exists: {p}. Trying to run the command. ")
                    if file_out: self.print(f"Creating directory {p.mkdir()}... ")
        if self.verbose: os.system(command)
        else: return os.popen(command).read()

ByteSize_class = {}
def ByteSize(x, format=' =9.04f'):
    T = type(x)
    if T in ByteSize_class: return ByteSize_class[T](x, format=format)
    class ByteSizeClass(T):
        units = " K M G T P E Z Y".split(' ')
        def __new__(cls, x, format=' =9.04f'):
            self = super().__new__(cls, x)
            self.format = format
            return self
        def __str__(self):
            i = int(math.log2(max(self, 1)) // 10)
            u = ByteSizeClass.units[i] + 'B'
            if len(u) == 1: u += ' '
            x = float(self / (1 << (10 * i)))
            return ("{x:%s} {u}"%self.format).format(x=x, u=u)
    ByteSize_class[T] = ByteSizeClass
    return ByteSizeClass(x, format=format)
